"""Commands that use AMYCNE python cli"""


def get_gctab_cmd(
    singularity: str, reference: str, binsize: str, path_gc_tab: str
) -> str:
    """Get a string with command to run the AMYCNE gc correction program.

    This is to identify problematic regions to exclude for AMYCNE fetal fraction and to
    do GC-correction
    """
    gc_tab_cmd = (
        f"{singularity} python /bin/AMYCNE/Generate_GC_tab.py "
        f"--fa {reference} --size {binsize} --n_mask > {path_gc_tab}"
    )

    return gc_tab_cmd


def run_amycne_cmd(
    singularity: str, out_prefix: str, path_gc_tab: str, minq: str
) -> str:
    """Get a string with command for running AMYCNEs fetal fraction

    This will predict the gender and how much dna is a sample is from the fetus.
    """
    amycne_cmd = (
        f"{singularity} python /bin/AMYCNE/AMYCNE.py --ff "
        f"--coverage {out_prefix}.tiddit.tab --gc {path_gc_tab} --Q {minq} > "
        f"{out_prefix}.tiddit.AMYCNE.tab"
    )
    return amycne_cmd



"""Commands that use bwa"""


def get_align_command(
    singularity: str, sample_id: str,tmp_dir: str, reference: str, fastq: str,out_dir:str, out_prefix: str, read: str = None
) -> str:
    """create a command to run bwa align. Read can be r1, r2 or none"""
    read_prefix = ".sai"
    if read == "r1":
        read_prefix = "_R1.sai"
    if read == "r2":
        read_prefix = "_R2.sai"

    fastqc_prefix=read_prefix.replace(".sai","")
    cmd = (
        f"{singularity} bwa aln -n 0 -k 0 {reference} {fastq} > "
        f"{out_prefix}{read_prefix}\n"
        f"{singularity} cat {fastq} | {singularity} fastqc stdin:{sample_id}{fastqc_prefix} -d {tmp_dir} -o {out_dir}/{sample_id}"
    )
    return cmd


def get_samse_command(
    singularity: str, reference: str, fastq: str, out_prefix: str
) -> str:
    """create a command for running bwa samsw"""
    cmd = f"{singularity} bwa samse -n -1 {reference} {out_prefix}.sai {fastq}"
    return cmd


def get_sampe_command(
    singularity: str, reference: str, fastq1: str, fastq2: str, out_prefix: str
) -> str:
    """create a command for running bwa sampe"""
    cmd = (
        f"{singularity} bwa sampe -n -1 {reference} "
        f"{out_prefix}_R1.sai {out_prefix}_R2.sai {fastq1} {fastq2}"
    )
    return cmd


def get_bamsormadup_command(
    singularity: str, tmp_dir: str, sample_id: str, out_prefix: str
) -> str:
    """Create a command for running bamorsamdup"""
    cmd = (
        f"{singularity} bamsormadup inputformat=sam threads=16 SO=coordinate "
        f"outputformat=bam tmpfile={tmp_dir}/{sample_id} "
        f"indexfilename={out_prefix}.bam.bai > {out_prefix}.bam"
    )
    return cmd



"""Code to print deliverables file"""

import logging
import yaml

from pathlib import Path
from typing import Iterator

from fluffy.version import __version__

LOG = logging.getLogger(__name__)

def print_deliverables(output_dir: Path,project_dir: Path, samples: list, project_id = "summary") -> None:
    """Create a deliverables YAML file"""

    deliverables={"files":[]}
    project_name=str(output_dir).strip("/").split("/")[-1]
    summary_path=output_dir.absolute() / f"{project_id}.csv"
    multiqc_path=output_dir.absolute() / "multiqc_report.html"

    deliverables["files"].append({"format":"csv", "id":project_name,"path":str(summary_path),"step":"summarise_batch","tag":"NIPT_csv"})
    deliverables["files"].append({"format":"html", "id":project_name,"path":str(multiqc_path),"step":"summarise_batch","tag":"MultiQC"})

    for sample in samples:
        sample_id=sample["sample_id"]
        zip_path=output_dir.absolute() / f"{sample_id}/{sample_id}.WCXpredict_aberrations.filt.bed"
        deliverables["files"].append({"format":"bed", "id":sample_id,"path":str(zip_path),"step":"summarise_batch","tag":"Wisecondor_aberrations"})

    f=open(f"{str(output_dir)}/deliverables.yaml","w")
    f.write(yaml.dump(deliverables))
    f.close()



"""CLI to rerun failed samples"""
import logging

import click

from fluffy.config import check_configs
from fluffy.deliverables import print_deliverables
from fluffy.status import print_status, check_status
from fluffy.workflows.analyse_samples import analyse_workflow
import shutil
import os

LOG = logging.getLogger(__name__)


@click.command()
@click.pass_context
@click.option(
    "--skip-preface", is_flag=True, help="Skip preface fetal fraction estimation"
)
@click.option("--dry-run", is_flag=True, help="Do not create any files")
def rerun(ctx, skip_preface, dry_run):
    """Run the pipeline to call NIPT"""
    LOG.info("Running fluffy rerun")
    samples = ctx.obj["samples"]
    configs = ctx.obj["configs"]
    configs["sample_sheet"] = ctx.obj["sample_sheet"]
    project_dir=ctx.obj["project"]
    slurm_api = ctx.obj["slurm_api"]

    config_path=configs["out"] / configs["name"]
    try:
        check_configs(configs, skip_preface=skip_preface)
    except FileNotFoundError as err:
        raise click.Abort
    output_dir=configs["out"]

    analysis_status = check_status(
        output_dir=output_dir
    )

    if analysis_status:
       run_version=analysis_status["fluffy_version"]
       analysis_time=analysis_status["analysis_time"]

    else:
       print("Error analysis status file is missing!")

    failed_samples=set([])

    first=True
    complete=True
    if analysis_status:
        for line in open(f"{output_dir}/sacct/fluffy_{analysis_time}.log.status"):
            content=line.strip().split()
            if first:
               first=False
               continue

            if not "COMPLETED" in content[-2]:
               complete=False

            if not "summarizebatch-" in line and not "COMPLETED" == content[-2] and not "fluffy-complete" in line:
          
               sample_id="-".join(content[1].split("-")[1:-1])
               print(f"FAIL:{content[1]}")
               failed_samples.add(sample_id)        

    if not analysis_status:
        pass       
    elif not analysis_status["analysis_run_status"] =="fail":
        print("Error: analysis status is not fail")
        print("make sure that the analysis is failed, and edit the analysis_run_status.json file")
        quit()

    if complete and analysis_status:
        print("Error: no failed jobs detected")
        print("make sure that the analysis is failed, or rerun it using the analysis module")
        quit()


    print_status(
        output_dir=output_dir,
    )

    for sample in failed_samples:
        if sample == "":
           continue

        try:
           shutil.rmtree(f"{output_dir}/{sample}")
        except:
           pass

        try:
           os.remove(f"{output_dir}/{sample}.fluffy-{run_version}.zip")
        except:
           pass

    if os.path.exists(f"{output_dir}/FAIL"):
        os.remove(f"{output_dir}/FAIL")

    if os.path.exists(f"{output_dir}/summary.csv"):
        os.remove(f"{output_dir}/summary.csv")

    if os.path.exists(f"{output_dir}/multiqc_report.html"):
        os.remove(f"{output_dir}/multiqc_report.html")

    try:
        shutil.rmtree(f"{output_dir}/multiqc_data")
    except:
        pass

    ok_samples=[]
    for sample in samples:
        if sample["sample_id"] in failed_samples:
            ok_samples.append(sample)

    analyse_workflow(
        samples=ok_samples,
        configs=configs,
        skip_preface=skip_preface,
        slurm_api=slurm_api,
        dry_run=dry_run,
    )



"""Commands to communicate with tiddit"""


def get_tiddit_cmd(singularity: str, out_prefix: str, binsize: str):
    """Return a string for command to calculate bins containing coverage values
    over the genome. This is done with Tiddit
    """
    tiddit_cmd = (
        f"{singularity} python /bin/TIDDIT.py --cov --bam {out_prefix}.bam "
        f" -z {binsize} -o {out_prefix}.tiddit"
    )

    return tiddit_cmd



"""Workflows to run picard"""
import logging

from fluffy.commands.pipelines import picard_qc
from fluffy.slurm_api import SlurmAPI

LOG = logging.getLogger(__name__)


def picard_qc_workflow(
    configs: dict,
    sample_id: str,
    afterok: int,
    slurm_api: SlurmAPI,
    dry_run: bool = False,
):
    """Run the picard pipeline"""
    LOG.info("Running the picard tools QC workflow")
    out_dir = configs["out"]
    picard_qc_pipe = picard_qc(configs=configs, out_dir=out_dir, sample_id=sample_id)

    jobid = slurm_api.run_job(
        name=f"picard_qc-{sample_id}",
        command=picard_qc_pipe,
        afterok=[afterok],
        dry_run=dry_run,
    )

    return jobid



"""Worklflows to run wisecondor"""

from typing import Iterator

from fluffy.commands.pipelines import wisecondor_x_test
from fluffy.commands.wisecondor import get_mkref_cmd
from fluffy.singularity_cmd import singularity_base
from fluffy.slurm_api import SlurmAPI
from fluffy.workflows.align import align_individual
from fluffy.workflows.status_update import pipe_complete
from fluffy.workflows.status_update import pipe_fail


def make_reference(
    samples: Iterator[dict], configs: dict, slurm_api: SlurmAPI, dry_run: bool = None
) -> int:
    """Create a reference based on some samples"""
    out_dir = configs["out"]
    jobids = []
    for sample in samples:
        

        sample_id = sample["sample_id"]
        sample_outdir = configs["out"] / sample_id

        # This will fail if dir already exists
        sample_outdir.mkdir(parents=True)

        slurm_api.slurm_settings["ntasks"]=configs["align"]["ntasks"]
        slurm_api.slurm_settings["mem"]=configs["align"]["mem"]

        align_jobid = align_individual(
            configs=configs, sample=sample, slurm_api=slurm_api, dry_run=dry_run
        )
        jobids.append(align_jobid)

        slurm_api.slurm_settings["ntasks"]=configs["slurm"]["ntasks"]
        slurm_api.slurm_settings["mem"]=configs["slurm"]["mem"]

    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])

    mkref_cmd = get_mkref_cmd(
        singularity=singularity,
        out=str(out_dir),
        testbinsize=configs["wisecondorx"]["testbinsize"],
        prefacebinsize=configs["wisecondorx"]["prefacebinsize"],
    )

    jobid = slurm_api.run_job(
        name="wcxmkref", command=mkref_cmd, afterok=jobids, dry_run=dry_run,
    )

    slurm_api.slurm_settings["time"]="1:00:00"
    pipe_complete(
        configs=configs, afterok=jobid, slurm_api=slurm_api, dry_run=dry_run
    )
    pipe_fail(
	configs=configs, slurm_api=slurm_api, dry_run=dry_run, afternotok=jobid
    )



    return jobid


def wisecondor_xtest_workflow(
    configs: dict,
    sample_id: str,
    afterok: int,
    slurm_api: SlurmAPI,
    dry_run: bool = False,
):
    """Run the wisecondor x test workflow"""
    out_dir = configs["out"]
    run_wcx_pipe = wisecondor_x_test(
        configs=configs, out_dir=out_dir, sample_id=sample_id
    )

    jobid = slurm_api.run_job(
        name=f"wcx-{sample_id}",
        command=run_wcx_pipe,
        afterok=[afterok],
        dry_run=dry_run,
    )

    return jobid



"""Pipeline commands that uses multiple tools"""

from pathlib import Path

from .amycne import get_gctab_cmd, run_amycne_cmd
from .bwa import (get_align_command, get_bamsormadup_command,
                  get_sampe_command, get_samse_command)
from fluffy.singularity_cmd import singularity_base
from .picard import (get_collect_gc_bias_cmd, get_collect_insert_size_cmd,
                     get_estimate_complexity_cmd)
from .tiddit import get_tiddit_cmd
from .utils import get_outprefix
from .wisecondor import get_convert_cmd, get_gender_cmd, get_predict_cmd


def align_and_convert_single_end(
    config: dict, fastq: str, out: Path, sample_id: str
) -> str:
    """create a command for running bwa and wisecondorX convert (single end)"""

    singularity=singularity_base(config["singularity"], config["out"], config["project"], config["singularity_bind"])

    out_prefix = get_outprefix(out, sample_id)
    aln = get_align_command(
        singularity=singularity,
        sample_id=sample_id,
        tmp_dir=config["align"]["tmpdir"],
        reference=config["reference"],
        fastq=fastq,
        out_dir=str(out),
        out_prefix=out_prefix,
    )

    samse_cmd = get_samse_command(
	singularity=singularity,
        reference=config["reference"],
        fastq=fastq,
        out_prefix=out_prefix,
    )
    bamsormadup_cmd = get_bamsormadup_command(
        singularity=singularity,
        tmp_dir=config["align"]["tmpdir"],
        sample_id=sample_id,
        out_prefix=out_prefix,
    )

    samse = " | ".join([samse_cmd, bamsormadup_cmd])

    convert = get_convert_cmd(
        singularity=singularity, out_prefix=out_prefix
    )

    return "\n".join([aln, samse, convert])


def align_and_convert_paired_end(
    config: dict, fastq: list, out: Path, sample_id: str
) -> str:
    """create a command for running bwa and wisecondorX convert (paired end)"""

    singularity=singularity_base(config["singularity"], config["out"], config["project"], config["singularity_bind"])

    out_prefix = get_outprefix(out, sample_id)
    aln_r1 = get_align_command(
        singularity=singularity,
       	sample_id=sample_id,
       	tmp_dir=config["align"]["tmpdir"],
        reference=config["reference"],
        fastq=fastq[0],
       	out_dir=str(out),
        out_prefix=out_prefix,
        read="r1",
    )

    aln_r2 = get_align_command(
        singularity=singularity,
       	sample_id=sample_id,
       	tmp_dir=config["align"]["tmpdir"],
        reference=config["reference"],
        fastq=fastq[1],
       	out_dir=str(out),
        out_prefix=out_prefix,
        read="r2",
    )

    sampe_cmd = get_sampe_command(
        singularity=singularity,
        reference=config["reference"],
        fastq1=fastq[0],
        fastq2=fastq[1],
        out_prefix=out_prefix,
    )

    bamsormadup_cmd = get_bamsormadup_command(
        singularity=singularity,
        tmp_dir=config["align"]["tmpdir"],
        sample_id=sample_id,
        out_prefix=out_prefix,
    )

    sampe = " | ".join([sampe_cmd, bamsormadup_cmd])

    convert = get_convert_cmd(
        singularity=singularity, out_prefix=out_prefix
    )

    return "\n".join([aln_r1, aln_r2, sampe, convert])


def amycne_ffy(configs: dict, out_dir: Path, sample_id: str) -> str:
    """fetal fraction estimation using tiddit and AMYCNE"""
    out_prefix = out_dir / sample_id / sample_id
    path_gc_tab = out_dir / sample_id / ".".join([sample_id, "gc.tab"])

    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])

    # Calculate coverage bins with tiddit
    tiddit_cmd = get_tiddit_cmd(
        singularity=singularity,
        out_prefix=str(out_prefix),
        binsize=configs["tiddit"]["binsize"],
    )

    # Calculate bins with GC and quality filtering
    gc_tab_cmd = get_gctab_cmd(
        singularity=singularity,
        reference=configs["reference"],
        binsize=configs["tiddit"]["binsize"],
        path_gc_tab=str(path_gc_tab),
    )

    amycne_cmd = run_amycne_cmd(
        singularity=singularity,
        out_prefix=str(out_prefix),
        path_gc_tab=str(path_gc_tab),
        minq=configs["amycne"]["minq"],
    )
    return "\n".join([tiddit_cmd, gc_tab_cmd, amycne_cmd])


def picard_qc(configs: dict, out_dir: Path, sample_id: str) -> str:
    """Get a string with pipeline steps to run picard qc"""
    out_prefix = out_dir / sample_id / sample_id
    reference = configs["reference"]
    javasettings = configs["picard"]["javasettings"]

    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])

    gc_bias_cmd = get_collect_gc_bias_cmd(
        singularity=singularity,
        out_prefix=str(out_prefix),
        reference=reference,
        javasettings=javasettings,
        tmp_dir=str(configs["align"]["tmpdir"]),
    )

    insert_size_cmd = get_collect_insert_size_cmd(
        singularity=singularity,
        out_prefix=str(out_prefix),
        javasettings=javasettings,
        tmp_dir=str(configs["align"]["tmpdir"]),
    )

    estimate_complexity_cmd = get_estimate_complexity_cmd(
        singularity=singularity,
        out_prefix=str(out_prefix),
        javasettings=javasettings,
        tmp_dir=str(configs["align"]["tmpdir"]),
    )

    return "\n".join([gc_bias_cmd, insert_size_cmd, estimate_complexity_cmd])


def wisecondor_x_test(configs: dict, out_dir: Path, sample_id: str) -> str:
    """Get the commands for running the wisecondor chromosome X test"""
    out_prefix = out_dir / sample_id / sample_id
    singularity=singularity_base(configs["singularity"], configs["out"], configs["project"], configs["singularity_bind"])
    blacklist = configs["wisecondorx"]["blacklist"]
    zscore = str(configs["wisecondorx"]["zscore"])

    wisecondor_test_cmd = get_predict_cmd(
        singularity=singularity,
        out_prefix=out_prefix,
        reference=configs["wisecondorx"]["reftest"],
        blacklist=blacklist,
        zscore=zscore,
    )

    wisecondor_preface_cmd = get_predict_cmd(
        singularity=singularity,
        out_prefix=out_prefix,
        reference=configs["wisecondorx"]["refpreface"],
        blacklist=blacklist,
        zscore=zscore,
        preface=True,
    )

    wisecondor_gender_cmd = get_gender_cmd(
        singularity=singularity,
        out_prefix=out_prefix,
        reference=configs["wisecondorx"]["reftest"],
    )

    return "\n".join(
        [wisecondor_test_cmd, wisecondor_preface_cmd, wisecondor_gender_cmd]
    )



"""Based on https://github.com/kennethreitz/setup.py"""

import io
import os
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup

# Package meta-data.
NAME = "cg-fluffy"
DESCRIPTION = "NIPT analysis pipeline"
URL = "https://github.com/Clinical-Genomics/fluffy"
EMAIL = "jesper.eisfeldt@scilifelab.com"
AUTHOR = "Jesper Eisfeldt"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = None

HERE = os.path.abspath(os.path.dirname(__file__))


def parse_reqs(req_path="./requirements.txt"):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with io.open(os.path.join(HERE, req_path), encoding="utf-8") as handle:
        # remove comments and empty lines
        lines = (
            line.strip() for line in handle if line.strip() and not line.startswith("#")
        )

        for line in lines:
            # check for nested requirements files
            if line.startswith("-r"):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])

            else:
                # add the line as a new requirement
                install_requires.append(line)

    return install_requires


# What packages are required for this module to be executed?
REQUIRED = parse_reqs()

# What packages are optional?
EXTRAS = {}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!


# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
ABOUT = {}
if not VERSION:
    with open(os.path.join(HERE, "fluffy", "version.py")) as f:
        exec(f.read(), ABOUT)
else:
    ABOUT["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(HERE, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(ABOUT["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=ABOUT["__version__"],
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": ["fluffy = fluffy.__main__:base_command"]},
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    keywords=["vcf", "compression"],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Intended Audience :: Science/Research",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)
